from kblackjack.card import Card, Suit
from kblackjack.players import Player, FakePlayer, RealPlayer
import random

def createDeck() -> list[Card]:
    deck = []
    for _, suit in Suit.__members__.items():
        # for ranks 2 through 10 + J, Q, K, A
        for rank in [str(x) for x in range(2, 11)] + list('JQKA'):
            deck.append(Card(rank, suit))
    random.shuffle(deck)
    return deck

def dealCards(deck: list[Card], num: int, players: list[Player]) -> None:
    """
    Deal the given number of cards to each player
    """
    for p in players:
        for _ in range(num):
            if not deck:
                print('\n\nDeck is empty! Ending game.')
                exit(2)
            p.cards.append(deck.pop())

def discardHands(discardList: list[Card], players: list[Player]) -> None:
    """
    Move all cards from all players' hands to the discard pile
    """
    for p in players:
        discardList.extend(p.cards)
        p.cards = []

def checkForBlackjack(players: list[Player]) -> list[Player]:
    result = []
    for p in players:
        if p.handTotal() == 21:
            result.append(p)
    return result

def doRound(players: list[Player], deck: list[Card], discarded: list[Card]) -> None:
    allPlayersStand = False

    # If any players got blackjack, print the winners and end the round
    winners = checkForBlackjack(players)
    if winners:
        print(f'Round over! {len(winners)} players got blackjack!')
        print('Winners:')
        for winner in winners:
            winner.roundsWon += 1
            print('\t' + winner.reprWithCards())
        return

    while not allPlayersStand:
        for player in players:
            print(player)

        # Allow each player to play their round
        allPlayersStand = True
        for player in players:
            # Dealer doesn't hit until the end
            if isinstance(player, FakePlayer) and player.isDealer:
                continue

            if player.willHit():
                allPlayersStand = False
                dealCards(deck, 1, [player])
                if isinstance(player, RealPlayer):
                    print(f'You received: {player.cards[-1]}')
                    print(f'You now have: {player.cards}')

        print('\n')

    # Locate dealer in players list
    dealer = next(filter(lambda p: isinstance(p, FakePlayer) and p.isDealer, players))
    while dealer.willHit():
        dealCards(deck, 1, [dealer])

    # At this point, all players have stood or busted, including the dealer

    # Find the winner(s)
    bestScore = 0
    winners = []
    for player in players:
        # Print cards
        print(player.reprWithCards())

        # Check if player is winner
        score = player.handTotal()
        if score == bestScore:
            winners.append(player)
        elif score > bestScore and score <= 21:
            bestScore = score
            winners = [player]

    print(f'\nRound over! Best score: {bestScore}')
    print('Winners:')
    for winner in winners:
        print('\t' + winner.reprWithCards())

def main():
    deck = createDeck()
    discarded = []
    players = []

    # Dealer is added first so they're printed first during rounds
    players.append(FakePlayer(isDealer=True))

    try:
        numPlayers = int(input('How many fake players would you like? '))
    except ValueError:
        print('Invalid number of players')
        exit(1)
    for i in range(numPlayers):
        players.append(FakePlayer())

    # Real player gets added last so they get printed at the bottom during
    # rounds
    players.append(RealPlayer())

    try:
        rounds = int(input('How many rounds would you like to play? '))
    except ValueError:
        print('Invalid number of players')
        exit(1)

    for i in range(rounds):
        dealCards(deck, 2, players)
        print(f'======= ROUND {i + 1} =======\n')
        doRound(players, deck, discarded)
        discardHands(discarded, players)
        if i + 1 < rounds:
            print('\n\n\n\n')



import itertools
from kblackjack.card import Card

class Player:
    _idCounter = itertools.count()

    id: int
    cards: list[Card]
    roundsWon: int

    def __init__(self, cards: list[Card] = None):
        if cards is None:
            cards = []
        self.cards = cards
        self.id = next(Player._idCounter)
        self.roundsWon = 0

    def __repr__(self) -> str:
        bust = ''
        if self.handTotal() > 21:
            bust = ' (BUST)'
        return f'Player {self.id}{bust}: {len(self.cards)} cards'

    def reprWithCards(self) -> str:
        bust = ''
        if self.handTotal() > 21:
            bust = ' (BUST)'
        return f'Player {self.id}{bust}: {self.cards}'

    def handTotal(self) -> int:
        """
        Return the highest sum of the given cards that doesn't exceed 21. If
        staying below 21 isn't possible, return the smallest possible sum.
        """
        aceCount = self.cards.count(Card('A'))
        sumWithoutAces = 0
        for c in self.cards:
            if c.rank != 'A':
                sumWithoutAces += c.value
        gap = 21 - sumWithoutAces
        if gap >= 11 + aceCount - 1 and aceCount >= 1:
            return sumWithoutAces + 11 + aceCount - 1
        else:
            return sumWithoutAces + aceCount

class FakePlayer(Player):
    def __init__(self, isDealer: bool = False):
        super().__init__()
        self.isDealer = isDealer

    def __repr__(self) -> str:
        bust = ''
        if self.handTotal() > 21:
            bust = ' (BUST)'
        if self.isDealer:
            return f'Dealer{bust}: {len(self.cards)} cards\n\tVisible card: {self.cards[1]}'
        else:
            return 'Fake ' + super().__repr__()

    def reprWithCards(self) -> str:
        if self.isDealer:
            name = 'Dealer'
        else:
            name = f'Fake Player {self.id}'
        return f'{name}: {self.cards}'

    def willHit(self) -> bool:
        """
        Simulates the player taking a turn. Currently will hit on 16 or below
        and stand on 17 and above.
        """
        return self.handTotal() < 16

class RealPlayer(Player):
    def __repr__(self) -> str:
        bust = ''
        if self.handTotal() > 21:
            bust = ' (BUST)'
        return f'You{bust}: {self.cards}'

    reprWithCards = __repr__

    def willHit(self) -> bool:
        if self.handTotal() > 21:
            return False

        choice = input('\nHit? (Y/n) ').lower()
        if choice[0] == 'n':
            return False
        elif not choice or choice[0] == 'y':
            return True
        else: # if not 'y*', 'n*', or empty, ask again
            return self.willHit()



from dataclasses import dataclass, field
from enum import Enum, unique

@unique
class Suit(Enum):
    SPADE = '\N{BLACK SPADE SUIT}'
    CLUB = '\N{BLACK CLUB SUIT}'
    HEART = '\N{BLACK HEART SUIT}'
    DIAMOND = '\N{BLACK DIAMOND SUIT}'

@dataclass(frozen=True, order=True)
class Card:
    rank: str
    suit: Suit = field(default=Suit.SPADE, compare=False)

    def __repr__(self):
        return f'{self.rank} {self.suit.value}'

    @property
    def value(self):
        if self.rank.isnumeric():
            return int(self.rank)
        elif self.rank == 'A':
            return 11
        return 10
